#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <math.h>

using namespace std;
#pragma warning(disable : 4996)

#define BUFFER_LENGTH (2048)
#define PHONEBUFFER_SIZE (10000)
#define SLOTSIZE (16)
long SHIFT = 0;

#define ByteSwap(x) ByteSwap5((unsigned char *) &x,sizeof(x))

void ByteSwap5(unsigned char * b, int n)
{
   register int i = 0;
   register int j = n-1;
   while (i<j)
   {
      std::swap(b[i], b[j]);
      i++, j--;
   }
}

int phone2id(char* temp, char *table, int *numstates, int *stateind){

	char *slot = table;
	int i=0;
	*stateind=0;
	while(i<PHONEBUFFER_SIZE && strcmp(temp, slot)){
		slot+=SLOTSIZE; 
		*stateind+=numstates[i];
		i++; 
	}
	if(i==PHONEBUFFER_SIZE) return -1;
	else return i;
}

int LoadPhoneSet(char *fname, char *table, int *numstates){

	char temp[SLOTSIZE+1];
	char *slot = table;
	int numphones = 0;
	char *ch;
	FILE* phonefile = fopen(fname, "rb");
	if(!phonefile){printf("Cannot open phone file %s\n!", fname); getchar(); return -1;}
	int R;
	do{
		R=fscanf(phonefile, "%s", &temp);
		if(R==EOF) break;
		ch = strchr(temp, '_');
		numstates[numphones] = 1;//if no state info, we assume 1 state per phone
		/*if(ch != NULL){ 
			numstates[numphones] = atoi(ch+1);
			if(!(numstates[numphones]==1 || numstates[numphones]==3)) printf("ERROR: The number of states must be 1 or 3!!!");
			*ch = '\0';
		}*/
		strncpy(slot, (char *)&temp, SLOTSIZE);
		slot += SLOTSIZE;
		numphones++;
	}while(R!=EOF && numphones<PHONEBUFFER_SIZE);
	return numphones;
}

void printline(FILE *outfile, float *data, int numFeat, int format, long sentenceindex, long frameindex, int phoneindex){

	float f;
	int i;

	switch(format){
	case 0: //fprintf(outfile, "%ld %ld %d\n", sentenceindex, frameindex, phoneindex); break;
			fwrite(&phoneindex, sizeof(int), 1, outfile); break;
	case 1: fprintf(outfile, "%ld %ld ", sentenceindex, frameindex);
			for(i=0; i<numFeat; i++){
				f=data[i];
				fprintf(outfile, "%f ", f);
			}
			fprintf(outfile, "%d\n", phoneindex);
			break;
	case 2: for(i=0; i<numFeat; i++){
				f=data[i];
				fprintf(outfile, "%f,", f);
			}
			fprintf(outfile, "%d\n", phoneindex+1);
			break;
	case 3: fprintf(outfile, "%d\n", phoneindex+1);
			break;
	case 4: f = 0.0;
			for(i=0; i<numFeat; i++) data[i] = f * expf(data[i]) + (1.0f-f)*(i==phoneindex ? 1.0f : 0.0f);
			fwrite(data, sizeof(float), numFeat, outfile); break;
			break;
	}
}

long GetNextEndtime(FILE *recfile, char* label, long sampPeriod){ 

	int R = fscanf(recfile, "%s", label); //we assume this equals the previous endtime
	if(R==EOF) return -999;
	fscanf(recfile, "%s", label);
	double rectd = atof(label) - 50000; //bedrotozzuk a recfile SHIFTjet...
	long rect = long(0.5+rectd/double(sampPeriod));
	fscanf(recfile, "%s", label);
	return rect;
}

int main(int argc, char *argv[]){

	char buffer[BUFFER_LENGTH];
	int FORMAT;
	int AVERAGE = 0;
	int BIGENDIAN = 0;
	char temp[BUFFER_LENGTH];
	long   nSamples;       
	long   sampPeriod;
	short  sampSize;
	short  parmKind; 
	long i, j, k;

	if((argc<8) || (argc>9)){
		printf("This program converts HTK lab files to (pre-)pfile format.\n");
		printf("It also requires the HTK mfc files for this!\n");
		printf("Parameters: lab-filelist mfc-path outfile phonelist format average bigendian [shift]\n");
		printf("Format can be: l: idx format, only the labels get saved\n");
		printf("               p: pre-pfile format, labels and features are both saved\n");
		printf("               c: a c4.5 style data file is saved (class numbers starting from 1)\n");
		printf("               r: only the labels are saved, and are taken from .rec files in the mfc path\n");
		printf("               ls: the labels are smoothed using the posteriors taken from the mfc\n");
		printf("(in the first 2 cases use pfile_create for conversion to pfile format)\n");
		printf("Average=0 means that each frames gets saved separately.\n");
		printf("Otherwise one feature vector is created per segment.\n");
		printf("Bigendian>0 will apply a byte order swap to the data.\n");
		printf("SHIFT>0 will shift the time indice of the frames. Say, 125000 will result in a\n");
		printf("half-frame shift if the frame size is 25ms\n");
		getchar();
		return -1;
	}
	char* listname = argv[1];
	char* mfcpath = argv[2];
	char* outname = argv[3];
	char* phonname = argv[4];
	if(!strcmp(argv[5], "l")) FORMAT = 0;
	else if(!strcmp(argv[5], "p")) FORMAT = 1;
	else if(!strcmp(argv[5], "c")) FORMAT = 2;
	else if(!strcmp(argv[5], "r")) FORMAT = 3;
	else if(!strcmp(argv[5], "ls")) FORMAT = 4;
	else {printf("Unknown format was specified$\n"); return -1;}
	if(atoi(argv[6])) AVERAGE = 1;
	if(atoi(argv[7])) BIGENDIAN = 1;
	if(argc==9) SHIFT = atoi(argv[8]);

	char phonetable[15*PHONEBUFFER_SIZE];
	int phonestates[PHONEBUFFER_SIZE];
	int numphones = LoadPhoneSet(phonname, phonetable, phonestates);
	if(numphones<=0) {printf("Cannot find phonefile! %s\n", phonname); getchar(); return -1;}
	printf("%i phones loaded from phonetable.\n", numphones);
//char spk[10];

	ifstream listFile(listname, ios::in/*|ios::nocreate*/);
	if(!listFile){printf("Cannot find listfile! %s\n", listname); getchar(); return -1;}

	FILE *outfile;
	if(FORMAT>0 && FORMAT<4) outfile = fopen(outname, "wt");
	else { //binary idx format
		if(FORMAT==4){
			outfile = fopen(outname, "wb");
			unsigned char ch = 0x0;
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			ch = 0x0D; //this codes float data type
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			ch = 2;
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			int empty = 0;
			fwrite(&empty, sizeof(int), 1, outfile);
			fwrite(&empty, sizeof(int), 1, outfile);
			}
		else{
			outfile = fopen(outname, "wb");
			unsigned char ch = 0x0;
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			ch = 0x0C; //this codes int data type
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			ch = 1;
			fwrite(&ch, sizeof(unsigned char), 1, outfile);
			int empty = 0;
			fwrite(&empty, sizeof(int), 1, outfile);
		}
	}

	long sentenceindex = 0;
	long framesum = 0;
	long fileframesum = 0;
	long segmentsum = 0;
	int numFeat;
	while (listFile && listFile.getline(buffer, BUFFER_LENGTH)){

		printf("Processing file %s\n", buffer);
		char *dot = strrchr(buffer, '.');
		if(dot) strcpy(dot, ".lab"); else strcat(buffer, ".lab");
		FILE* labfile = fopen(buffer, "rt");
		if(!labfile){printf("Cannot open lab file %s\n!", buffer); getchar(); return -1;}
		dot = strrchr(buffer, '.');
		if(dot) strcpy(dot, ".mfc");
		strcpy(temp, mfcpath);
		dot = strrchr(buffer, '/');
		if(!dot) dot = strrchr(buffer, '\\');
		if(!dot) dot = buffer-1;
		if(mfcpath[strlen(mfcpath)-1]!='/' && mfcpath[strlen(mfcpath)-1]!='\\') strcat(temp, "/");
		strcat(temp, dot+1);
//strncpy(spk, dot+1, 5); //this is to dig out timit speakers
		FILE* mfcfile = fopen(temp, "rb");
		if(!mfcfile){printf("Cannot open mfc file %s\n!", temp); fclose(labfile); getchar(); return -1;}
		strcpy(temp+strlen(temp)-4, ".rec");
		FILE* recfile = NULL;
		if(FORMAT==3) recfile = fopen(temp, "rt");
		if(FORMAT==3 && !recfile){printf("Cannot open rec file %s\n!", temp); fclose(labfile); fclose(mfcfile); getchar(); return -1;}
		fread(&nSamples, sizeof(long),  1, mfcfile);
		fread(&sampPeriod, sizeof(long),  1, mfcfile);
		fread(&sampSize, sizeof(short), 1, mfcfile);
		fread(&parmKind, sizeof(short), 1, mfcfile);
		if(BIGENDIAN){
		        ByteSwap(nSamples);
                ByteSwap(sampPeriod);
                ByteSwap(sampSize);
                ByteSwap(parmKind);
		}
//		sampPeriod = 100000;
		numFeat = sampSize / sizeof(float);
		printf("Number of frames: %ld\n", nSamples);
		fileframesum += nSamples;
		//loading mfc file
		float *mfcdata = new float[numFeat * (nSamples+5) + 1];
		float *avgs = mfcdata + numFeat*nSamples;
		float f;
		float *fp = mfcdata;
		for(i=0; i<numFeat*nSamples; i++){
				fread(&f, sizeof(float), 1, mfcfile);
				if(BIGENDIAN) ByteSwap(f);
				*fp = f;
				fp++;
		}
		fclose(mfcfile);

		int phoneid = -1;
		int guessid = -1;
		int stateind;
		long frame = 0;
		long t1, t2, tc1, tc2, tn1, tn2;
		double t1d, t2d, tc1d, tc2d;
		long rect = 0;
		int f1seg = 0;
		if(FORMAT==3){
			rect = GetNextEndtime(recfile, (char *)&temp, sampPeriod);
			guessid = phone2id(temp, phonetable, phonestates, &stateind);
			if(guessid==-1) printf("WARNING: unknown phoneme %s\n found!\n", temp);
		}
		t2 = 0;
		int R = fscanf(labfile, "%s", &temp);
		while(R!=EOF){
				t1d = atof(temp) - SHIFT;
				t1 = long(0.50003+t1d/double(sampPeriod));
				if(f1seg) t1++; //the previous segment had only one frame
				if(t1<0){ printf("WARNING: segment start time < 0! (%d)\n", t1); t1 = 0; }
				else if(t1>=nSamples){ printf("ERROR: segment start time > nSamples! (%d>%d)\n", t1, nSamples); t1 = nSamples-1; }
				//the next line is required if we want to process all frames
				if(frame==0 && t1>0) { printf("WARNING: %ld unlabeled frames added to first segment!\n", t1); t1 = 0; }
				//checking if there was an unlabeled segment
				if(t2 != t1) printf("WARNING: %ld unlabeled frames were found!\n", t1-t2);
				fscanf(labfile, "%s", &temp);
				t2d = atof(temp) - SHIFT;
				t2 = long(0.50003+t2d/double(sampPeriod));
				if(t2<0){ printf("ERROR: segment end time < 0!\n"); t2 = 0; }
				else if(t2>nSamples){ printf("WARNING: segment end time > nSamples!\n");  t2 = nSamples; } // lehet, mert j<t2-ig megy a ciklus!
				fscanf(labfile, "%s", &temp);
				phoneid = phone2id(temp, phonetable, phonestates, &stateind);
//				phoneid = phone2id(spk, phonetable, phonestates, &stateind); //--this is for speaker labels
				if(phoneid==-1) printf("WARNING: unknown phoneme %s\n found!\n", temp);
				R = fscanf(labfile, "%s", &temp);
				//the next line is required if we want to process all frames
				if(R == EOF && t2<nSamples) { printf("WARNING: %ld unlabeled frames added to last segment!\n", nSamples-t2); t2 = nSamples; }
				if(t2==t1) { printf("WARNING: 0-frame segment found!\n"); t2++; f1seg = 1;} //to save something even for 0-frame segments
				else f1seg = 0;
				segmentsum++;
				//frame-oriented saving
				if(!AVERAGE){
					//printf("%d %d\n", t1, t2);
					if(phonestates[phoneid]==1){ //1 state
						if(FORMAT==3){
							for(int i=t1; i<t2; i++){
								while(i>=rect){
									rect = GetNextEndtime(recfile, (char *)&temp, sampPeriod);
									if(rect==-999) break;
									guessid = phone2id(temp, phonetable, phonestates, &stateind);
									if(guessid==-1) printf("WARNING: unknown phoneme %s\n found!\n", temp);
								}
								printline(outfile, mfcdata+i*numFeat, numFeat, FORMAT, sentenceindex, frame, guessid);
								frame++;
							}
						}
						else for(int i=t1; i<t2; i++){
							printline(outfile, mfcdata+i*numFeat, numFeat, FORMAT, sentenceindex, frame, stateind);
							frame++;
						}
					}
					else if(phonestates[phoneid]==3){ //3 states
						double dur = t2d-t1d;
						tc1d = t1d + 0.3 * dur;
						tc2d = t1d + 0.7 * dur;
						tc1 = long(0.5+tc1d/double(sampPeriod));
						tc2 = long(0.5+tc2d/double(sampPeriod));
						if(tc1<0) tc1 = 0;
						else if(tc1>=nSamples) tc1 = nSamples-1;
						if(tc2<0) tc2 = 0;
						else if(tc2>=nSamples) tc2 = nSamples;
						for(int i=t1; i<tc1; frame++, i++)  printline(outfile, mfcdata+i*numFeat, numFeat, FORMAT, sentenceindex, frame, stateind);
						for(int i=tc1; i<tc2; frame++, i++) printline(outfile, mfcdata+i*numFeat, numFeat, FORMAT, sentenceindex, frame, stateind+1);
						for(int i=tc2; i<t2; frame++, i++)  printline(outfile, mfcdata+i*numFeat, numFeat, FORMAT, sentenceindex, frame, stateind+2);
					}
				}
				else{ //averaging over segments
					double dur = t2d-t1d;
					tc1d = t1d + 0.3 * dur;
					tc2d = t1d + 0.7 * dur;
					tc1 = long(0.5+tc1d/double(sampPeriod));
					tc2 = long(0.5+tc2d/double(sampPeriod));
					tn1 = t1 - 4;
					tn2 = t2 + 4;
					if(tn1<0) tn1 = 0;
					else if(tn1>=nSamples) tn1 = nSamples-1;
					if(t1<0) t1 = 0;
					else if(t1>=nSamples) t1 = nSamples-1;
					if(tc1<0) tc1 = 0;
					else if(tc1>=nSamples) tc1 = nSamples-1;
					if(tc2<0) tc2 = 0;
					else if(tc2>=nSamples) tc2 = nSamples-1;
					if(t2<0) t2 = 0;
					else if(t2>=nSamples) t2 = nSamples-1;
					if(tn2<0) tn2 = 0;
					else if(tn2>=nSamples) tn2 = nSamples-1;
					float *fp = avgs;
					i = tn1;
					for(j=0; j<numFeat; j++) fp[j] = mfcdata[i*numFeat+j];
					i++; k = 1;
					while(i<t1){ for(j=0; j<numFeat; j++) fp[j] += mfcdata[i*numFeat+j]; i++; k++;}
					for(j=0; j<numFeat; j++) fp[j] /= k;
					fp += numFeat;
					for(j=0; j<numFeat; j++) fp[j] = mfcdata[i*numFeat+j];
					i++; k = 1;
					while(i<tc1){ for(j=0; j<numFeat; j++) fp[j] += mfcdata[i*numFeat+j]; i++; k++;}
					for(j=0; j<numFeat; j++) fp[j] /= k;
					fp += numFeat;
					for(j=0; j<numFeat; j++) fp[j] = mfcdata[i*numFeat+j];
					i++; k = 1;
					while(i<tc2){ for(j=0; j<numFeat; j++) fp[j] += mfcdata[i*numFeat+j]; i++; k++;}
					for(j=0; j<numFeat; j++) fp[j] /= k;
					fp += numFeat;
					for(j=0; j<numFeat; j++) fp[j] = mfcdata[i*numFeat+j];
					i++; k = 1;
					while(i<t2){ for(j=0; j<numFeat; j++) fp[j] += mfcdata[i*numFeat+j]; i++; k++;}
					for(j=0; j<numFeat; j++) fp[j] /= k;
					fp += numFeat;
					for(j=0; j<numFeat; j++) fp[j] = mfcdata[i*numFeat+j];
					i++; k = 1;
					while(i<tn2){ for(j=0; j<numFeat; j++) fp[j] += mfcdata[i*numFeat+j]; i++; k++;}
					for(j=0; j<numFeat; j++) fp[j] /= k;
					fp += numFeat;
					*fp = t2-t1; //duration feature
					if(t2-t1>100) printf("WARNING: Suspiciously long segment (%i)!\n", t2-t1);
					printline(outfile, avgs, 5*numFeat+1, FORMAT, sentenceindex, frame, phoneid);
					frame++; //in this case the number of segments in fact...
				}
		}
		delete[] mfcdata;
		fclose(labfile);
		if(recfile) fclose(recfile);
		framesum += frame;
		sentenceindex++;
		printf("Labelled frames: %ld (%ld)\n", frame, frame-nSamples);
//		if(framesum!=fileframesum)
//			printf("\n");
	}//while loop over filelist lines
	listFile.close();
	printf("Labelled %ld frames (%ld segments) in %ld sentences\n", framesum, segmentsum, sentenceindex);
	printf("Processed %ld frames\n", fileframesum);
	fclose(outfile);

	if(!FORMAT){ //correcting the header of the idx file
		outfile = fopen(outname, "rb+");
		unsigned char ch = 0x0;
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		ch = 0x0C; //this codes int data type
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		ch = 1;
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		int iii = (int)framesum;
		fwrite(&iii, sizeof(int), 1, outfile);
	}
	else if(FORMAT==4){
		outfile = fopen(outname, "rb+");
		unsigned char ch = 0x0;
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		ch = 0x0D; //this codes float data type
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		ch = 2; //there are two additional size fields
		fwrite(&ch, sizeof(unsigned char), 1, outfile);
		int iii = (int)framesum;
		fwrite(&iii, sizeof(int), 1, outfile);
		fwrite(&numFeat, sizeof(int), 1, outfile); 
		fclose(outfile);
	}

//	getchar();
	return 0;
}